#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include "typedefs.h"
#include "smalloc.h"
#include "fatal.h"
#include "futil.h"
#include "physics.h"
#include "vec.h"
#include "edsam.h"
#include "network.h"
#include "mdrun.h"
#include "repl_ex.h"

#ifdef GMX_MPI
#include "mpi.h"
#endif

#ifdef GMX_DOUBLE
#define GMX_MPI_REAL MPI_DOUBLE
#else
#define GMX_MPI_REAL MPI_FLOAT
#endif

#define MIN(a,b) ((a)<(b)) ? (a) : (b)


/*
 * Wrapper functions rvec[sub,add,norm,smul,*_to_one,copy] made static
 * to keep them 'private' within teerex.c & avoid confusion with GMX
 * rvec-routines (e.g. rvec_add <-> rvecadd).
 * Do the same with other helper functions.
 */ 
static inline void rvecsub(int dim,rvec *a, rvec *b, rvec *c) 
{
  /* c=a-b; */
  int i;
  for (i=0; i<dim; i++) 
    rvec_sub(a[i],b[i],c[i]);
}
  
static inline void rvecadd(int dim,rvec *a, rvec *b, rvec *c) 
{
  /* c=a+b; */
  int i;
  for (i=0; i<dim; i++)
    rvec_add(a[i],b[i],c[i]);
}

static inline real rvecnorm(int dim,rvec *a) 
{
  int  i;
  real sum=0;
  for (i=0; i<dim; i++)
    sum += norm2(a[i]);
  return sqrt(sum);
}

static inline void rvecsmul(int dim, real s, rvec *a) 
{
  int i;
  for (i=0; i<dim; i++)
    svmul(s,a[i],a[i]);
} 

static inline void rvec_to_one(int dim, rvec *a) 
{
  rvecsmul(dim,1.0/rvecnorm(dim,a),a);
}

static inline void rveccopy(int dim, rvec *a, rvec *b) 
{
  /*b=a;*/
  int i;
  for (i=0; i<dim; i++)
    copy_rvec(a[i],b[i]);
} 

static inline real calc_temp(real ekin,real nrdf)
{
  if (nrdf > 0)
    return (2.0*ekin)/(nrdf*BOLTZ);
  else
    return 0.0;
}

/*
 * main functions of TEE-REX module
 */
extern t_teerex *init_TEEREX(FILE *log,t_commrec *mcr,t_inputrec *ir,t_groups *grps,t_mdatoms *mdatoms,t_edsamyn *edyn,gmx_repl_ex_t *re)
{
  int          i, j;
  int          ierr;
  const int    master = 0;
  real         temp;            /* non-{es} group temperature of current replica        */
  real         *nes_temp;       /* for collecting temperatures of all non-{es} groups   */
  real         nrdf_sys;        /* #DOF of system                                       */
  real         dummy;         
  t_edpar      *edi;
  t_teerex     *teerex = NULL;

#ifdef GMX_DOUBLE               /* for MPI_Gather() */
#define mpi_type MPI_DOUBLE
#else
#define mpi_type MPI_FLOAT
#endif
  
  snew(teerex,1);
  snew(nes_temp,re->nrepl);

  edi = edyn->edpar;  

  if (MASTER(mcr))
    {
      fprintf(stderr,"Initializing TEE-REX module...\n");  
      fprintf(log,"TEEREX  Initializing TEE-REX...\n");  
    }
  teerex->bTEEREX    = TRUE;
  teerex->bidGrpTemp = FALSE;
  edyn->bEdsam       = FALSE;     /* switch off ED sampling during TEE-REX run to   */
			          /* avoid interference from ED routines.           */
  teerex->mcr        = mcr;       /* need mcr info for gmx_sum() in TEEREX_tcoupl() */
  teerex->mdatoms    = mdatoms;   /* need for T-scaling after exchange              */
  teerex->ir         = ir;
  teerex->grps       = grps;      

 
  /* Check of temperature coupling group entries of *.mdp:
   * Correct entries have
   *  - at least two temperature groups (ess.subspace {es}, non-{es} [nes])
   *  - {es}-group MUST be first T-group (b/c of replica exchange modules);
   *    ASSUMING it is so, I don't actually check it!
   *  - all nes-groups MUST have the same temperature among ALL replicas
   *  - temperature of each {es}-group MUST be equal or higher than the nes-group 
   *    temperatures (replica with T_es == T_nes is not T_es-coupled)
   *  
   * Example: 4 replicas
   *     (T_es-group)       (T_nes-groups)
   * repl  Backbone    Protein_&_!Backbone    Other
   *  0      319K            300K             300K
   *  1      300K            300K             300K  <- REFERENCE replica, not T_es coupled
   *  2      312K            300K             300K
   *  3      306K            300K             300K
   *
   * Temperatures don't have to be in order, init_replica_exchange() takes
   * care of the ordering.
   */

  /* at least two T-groups? */
  if (ir->opts.ngtc == 1)
    gmx_fatal(FARGS,"TEEREX:%i  At least two temperature groups required",re->repl);

  /*  T_es > T_0? (i.e. ref_t[0] >= ref_t[i>0]) */
  for (i=1; i<(ir->opts.ngtc); i++)
    if (ir->opts.ref_t[0] < ir->opts.ref_t[i])
      gmx_fatal(FARGS,"TEEREX:%i  T_es must be higher than T_0!",re->repl);

  /* all T_0 identical in one replica? */
  for (i=1; i<(ir->opts.ngtc); i++)
    if (ir->opts.ref_t[i] != ir->opts.ref_t[1])
      gmx_fatal(FARGS,"TEEREX:%i  T_0 not identical among non-{es} groups in replica %i",re->repl,re->repl);

  /* T_0 identical among replicas? */
  temp = ir->opts.ref_t[1];
  ierr = MPI_Gather(&temp,1,mpi_type,nes_temp,1,mpi_type,master,MPI_COMM_WORLD); /* collect ref_t[1] from replicas & store in nes_temp */
  if (ierr != 0)
    gmx_fatal(FARGS,"TEEREX:%i  MPI_Gather() failed. Cannot check non-{es} temperatures",re->repl);
  for (i=1; i<(re->nrepl); i++)       /* check after collecting T_nes from jobs */
    if (nes_temp[i] != nes_temp[0])
      gmx_fatal(FARGS,"TEEREX:%i  T_0 not identical among replicas",re->repl);
  sfree(nes_temp);

  /* no T_es-coupling if T_es == T_nes */
  teerex->bidGrpTemp = (re->temp[re->repl] == ir->opts.ref_t[1]);
  if (teerex->bidGrpTemp)
    fprintf(stderr,"NO T_es-coupling for reference replica (T_es=%.2f, T_0=%.2f)\n",re->temp[re->repl],ir->opts.ref_t[1]);
  

  /* Now prepare ED data for TEEREX_tcoupl() routine:
   * - silence ED output (edpar->outfrq, edpar->logfrq)
   *   and other vars (just to be safe)
   * - switch off origin & target structure, since we're not doing 
   *   targeted MD (edpar->s[ori|tar].nr=0). 
   *   Groups not containing evecs have edyn->edpar->vecs.*.neig = 0 
   *   (default from make_edi).
   * - move evec info (t_eigvec struct) into monitoring group
   *   (linfix, linacc, radfix, radacc, radcon --> mon) so I don't
   *   have to check location of evecs.
   */

  /* silence ED output & other vars */
  edi->outfrq     = 0;
  edi->logfrq     = 0;
  edi->presteps   = 0;
  edi->maxedsteps = 0;

  /* switch off origin & target structure */
  edi->sori.nr = 0;
  edi->star.nr = 0;

  /* collect evec info in monitoring group -> easy access in TEEREX_tcoupl() */
  if (edi->vecs.mon.neig > 0)
    {
      if (MASTER(mcr))
	fprintf(stderr,"Evecs already in monitoring group\n");
    }
  else if (edi->vecs.linfix.neig > 0)
    {
      edi->vecs.mon = edi->vecs.linfix;
      edi->vecs.linfix.neig = 0;
      if (MASTER(mcr))
	fprintf(stderr,"Moved evecs: linfix -> mon\n");
    }
  else if (edi->vecs.linacc.neig > 0)
    {
      edi->vecs.mon = edi->vecs.linacc;
      edi->vecs.linacc.neig = 0;
      if (MASTER(mcr))
	fprintf(stderr,"Moved evecs: linacc -> mon\n");
    }
  else if (edi->vecs.radfix.neig > 0)
    {
      edi->vecs.mon = edi->vecs.radfix;
      edi->vecs.radfix.neig = 0;
      if (MASTER(mcr))
	fprintf(stderr,"Moved evecs: radfix -> mon\n");
    }
  else if (edi->vecs.radacc.neig > 0)
    {
      edi->vecs.mon = edi->vecs.radacc;
      edi->vecs.radacc.neig = 0;
      if (MASTER(mcr))
	fprintf(stderr,"Moved evecs: radacc -> mon\n");
    }
  else if (edi->vecs.radcon.neig > 0)
    {
      edi->vecs.mon = edi->vecs.radcon;
      edi->vecs.radcon.neig = 0;
      if (MASTER(mcr))
	fprintf(stderr,"Moved evecs: radcon -> mon\n");
    }
  MPI_Barrier(MPI_COMM_WORLD);

  /* show coupled {es} evecs & exchange attempt frequency */
  if (!teerex->bidGrpTemp)
    {
      if (re->repl==1)
	fprintf(stderr,"Coupling %i eigenvector%s to T_es: ",
		edi->vecs.mon.neig,(edi->vecs.mon.neig > 1 ? "s" : ""));
      fprintf(log,"TEEREX:%i  coupling %i eigenvector%s to T_es[%i]: ",
	      re->repl,edi->vecs.mon.neig,(edi->vecs.mon.neig > 1 ? "s" : ""),re->repl);
      for (i=0; i<edi->vecs.mon.neig; i++)
	{
	  if (re->repl==1)
	    fprintf(stderr,"%i ",edi->vecs.mon.ieig[i]);
	  fprintf(log,"%i ",edi->vecs.mon.ieig[i]);
	}
      if (re->repl==1)
	fprintf(stderr,"\n");
      fprintf(log,"\n");
    }
  MPI_Barrier(MPI_COMM_WORLD);
  if (MASTER(mcr))
    fprintf(stderr,"Attempt exchanges every %isteps = %.4fps (dt = %.4fps)\n",
	    re->nst,(re->nst)*(ir->delta_t),ir->delta_t);
  fprintf(log,"TEEREX:%i  attempt exchanges every %i steps / %.4fps (dt = %.4fps)\n",
	      re->repl,re->nst,(re->nst)*(ir->delta_t),ir->delta_t);

  /* grep T_es from mdp file & communicate to all replicas */
  snew(teerex->T_es,re->nrepl);
  teerex->T_es[re->repl] = ir->opts.ref_t[0];      
  gmx_sum(re->nrepl,teerex->T_es,mcr);

  /* estimate effective SYSTEM temperature, used ONLY for TEE-REX (REMD) exchange criterion */
  snew(teerex->Tsys_eff,re->nrepl);
  nrdf_sys = 0.0;
  for (i=0; i<ir->opts.ngtc; i++)
    nrdf_sys += ir->opts.nrdf[i];  /* nrdf of total system */
  teerex->nrdf_sys = nrdf_sys;
  if (teerex->bidGrpTemp)  /* no Tsys_eff for reference replica */
    {
      teerex->Tsys_eff[re->repl] = ir->opts.ref_t[1];  /* needed in replica_exchange() */
      fprintf(log,"TEEREX:%i  REFERENCE replica, effective system temperature: Tsys_eff = %.2f\n",
	      re->repl,teerex->Tsys_eff[re->repl]);
    }
  else
    {
      /* Tsys_eff[repl m] = T_0 + (T_es[repl m] - T_0) * (N_es / N_df) */
      teerex->Tsys_eff[re->repl] = ir->opts.ref_t[1]+(teerex->T_es[re->repl]-ir->opts.ref_t[1])*(edi->vecs.mon.neig/nrdf_sys);
      fprintf(log,"TEEREX:%i  effective system temperature: Tsys_eff = %.2f   (T_es = %.2f)\n",
	      re->repl,teerex->Tsys_eff[re->repl],teerex->T_es[re->repl]);
    }
  gmx_sum(re->nrepl,teerex->Tsys_eff,mcr);  /* communicate to all replicas */

  /* init Berendsen T_es-coupling */
  teerex->lambda_es = 1.0;
  teerex->tau_es    = edi->flood.tau;

  /* write T-coupling parameters to log files */
  if (MASTER(mcr))
    fprintf(stderr,"Writing T-coupling parameters to log files\n");
  fprintf(log,"TEEREX:%i  T-coupling parameters ([T] = K, [tau_t] = ps):\n",re->repl);  
  if (!teerex->bidGrpTemp)
    fprintf(log,"TEEREX:%i  T[es] = %.2f,   tau_t[es] = %.4f\n",re->repl,teerex->T_es[re->repl],teerex->tau_es);
  for (i=0; i<(ir->opts.ngtc); i++)
    fprintf(log,"TEEREX:%i  T[%i] = %.2f,   tau_t[%i] = %.4f\n",re->repl,i,ir->opts.ref_t[i],i,ir->opts.tau_t[i]);

  /* no mass-weighting of projections */
  for (i=0; i<edi->sav.nr; i++)
    edi->sav.sqrtm[i] = 1.0;
  if (MASTER(mcr))
    fprintf(stderr,"Switched off mass-weighting in all replicas\n");
  fprintf(log,"TEEREX:%i  switched off mass-weighting in all replicas\n",re->repl);

  /* probably not necessary b/c already set outfrq & logfrq == 0 */
  /* no sam.edo output file(s) */
  ffclose(edi->edo);

#undef mpi_type

  return teerex;
}


extern void TEEREX_tcoupl(FILE *log,int step,rvec xprime[],t_edsamyn *edyn,
			  gmx_repl_ex_t *re,t_state *state,t_teerex *teerex,real fac) 
{
  int               i, j;
  int               ned;           /* index # of last atom belonging to index group */ 
  int               neig;          /* # evecs */
  int               nr;            /* # atoms of pca group (=index group) */
  static rvec       **es_evec;     /* evecs spanning essential subspace {es} */
  static rvec       *v_es;         /* {es} velocity projection written in cartesian basis */
  static rvec       *v_nes;        /* v_nes = v-v_es; written in cartesian basis */
  matrix            rotmat;
  real              mas, rad;
  bool              bFirst;        /* for initializing loc struct */
  real              dt, dt_1;
  struct t_do_edsam *loc;
  t_edpar           *edi;
  t_mdatoms         *md;
  t_inputrec        *ir;
  real              ekin;          /* Ekin of index atoms calc. with v projected into {es} <=> E_kin[{es}] */
  real              T;             /* current essential subspace temperature T_es[repl m] */
  real              dummy;          
  real              tau_es;        /* coupling constant for T_es */


  /***************************************************************/
  /* START   initialize everything (and more, uah,ha,ha)   START */
  /***************************************************************/
  ir     = teerex->ir;
  md     = teerex->mdatoms;
  edi    = edyn->edpar;             
  ned    = edi->ned;                 /* highest atom # of index group */
  neig   = edi->vecs.mon.neig;       /* # eigenvecs == dim{es}        */
  nr     = edi->sav.nr;              /* # atoms in index group; CAUTION: sref.nr & sav.nr can differ:    */
                                     /* sref.nr != sav.nr if reference structure to which ls-fit is done */  
				     /* differs from the T_es coupled structure.                         */
  dt     = ir->delta_t;
  dt_1   = 1.0/dt;
  tau_es = teerex->tau_es;           /* T-coupling constant for {es} */

  /* initialise loc struct necessary for fitting (vel. correction) */
  if(edi->local->do_edsam != NULL)
    {
      bFirst = FALSE;
      loc    = edi->local->do_edsam;
    }
  else
    {
      bFirst = TRUE;
      snew(edi->local->do_edsam,1);
    }
  loc = edi->local->do_edsam;

  if (bFirst) 
    {
      snew(loc->x,ned);
      snew(loc->xdum,ned);
      snew(loc->vdum,ned);
      snew(loc->vdum2,ned);
      snew(loc->fdum,ned);
      snew(loc->transvec,ned);
      snew(loc->old_transvec,ned);
      snew(loc->transvec_compact,ned);
      snew(loc->older_transvec,ned);
      snew(loc->vcorr,ned);
      snew(loc->old_vcorr,ned);

      /* save {es} evecs in **es_evec; needed for v_es & v_nes construction */
      snew(es_evec,neig);
      for (i=0; i<neig; i++)
	snew(es_evec[i],nr);
      for (i=0; i<neig; i++)
	es_evec[i] = edi->vecs.mon.vec[i];

      /* {es} velocity projections */
      snew(v_es,nr);
      snew(v_nes,nr);
    }
  /***************************************************************/
  /* END     initialize everything (and more, uah,ha,ha)     END */
  /***************************************************************/




  /* Tasks carried out by TEEREX_tcoupl():
   * (1) T_es coupling
   * (2) velocity scaling after successful exchange
   *
   * The variable 'fac' is used as a switch to distinguish tasks,  
   * fac  = -1    T_es coupling
   * fac != -1    v-scaling after successful replica exchange 
   */

  /*****************************************************************/
  /* START   fitting structure before projecting into {es}   START */
  /*****************************************************************/
  /* NB: xprime   = x(t+dt)   */
  /* NB: state->v = v(t+dt/2) */
  rveccopy(ned,xprime,loc->x);      /* working with copy of new & constrained coordinates (xprime) saves rmfit() */
  rveccopy(ned,state->v,loc->vdum); /* use copy of velocities for T_es coupling; copy back to state->v after coupling */
  rveccopy(ned,loc->x,loc->xdum);   /* get (second?) copy of positions */

  /* fit the structure */
  fitit(ned,loc->x,edi,loc->transvec,rotmat);

  rveccopy(ned,loc->transvec,loc->transvec_compact);

  /* this is to remove virtual jumps in translational velocity due to periodic boundaries */
  remove_pbc_effect(ned,loc->transvec_compact,state->box,edi); 

  if (bFirst || fac!=-1) 
    {
      copy_mat(rotmat,loc->old_rotmat);
      rveccopy(ned,loc->transvec_compact,loc->old_transvec);
      rveccopy(ned,loc->old_transvec,loc->older_transvec);
      dt_1 = 0;  /* no velocity correction for rot+trans */
    }

  /* correct velocities for the rotational and translational motion */
  rotate_vec(ned,loc->vdum,rotmat);                /* rotate velocities */
  rvecadd(ned,loc->xdum,loc->transvec,loc->xdum);  /* move copy of actual structure into origin */
  rotate_vec(ned,loc->xdum,loc->old_rotmat);       /* rotate positions with old matrix -> rotation in positions can be extracted */
  if ((step%re->nst) == 1)                         /* switch off velocity correction in first step after TEE-REX exchange */
    dt_1 = 0;
  correct_vel(ned,loc->vdum,loc->x,loc->xdum,-1.0*dt_1,loc->vcorr);

  /* bookkeeping for next integration step */
  rveccopy(ned,loc->vcorr,loc->old_vcorr);
  copy_mat(rotmat,loc->old_rotmat);
  rveccopy(ned,loc->old_transvec,loc->older_transvec);
  rveccopy(ned,loc->transvec_compact,loc->old_transvec);
  /*****************************************************************/
  /* END     fitting structure before projecting into {es}     END */
  /*****************************************************************/


  /* no T_es coupling for reference replica */
  if (teerex->bidGrpTemp && fac==-1)  
    return;


  /*********************************************************/
  /* START      T_es coupling / velocity scaling     START */
  /*********************************************************/
  /* PROCEDURE:
   * - create v_es vector (in t_teerex struct) form loc->vdum:
   *   v_es = sum{i in es} (loc->vdum)*evec[i]. 
   *   v_nes = v(cartesian)-v_es
   * - calculate lambda_es via Berendsen scheme => need T_es calculated for this!!!!!
   * - couple v_es with lambda_es 
   * - 'project' back to cartesian frame: v'(cartesian) = (v_es)'+(v_nes)'
   * - remove fitting of xprime(==loc->x) since we don't work on copy of coordinates
   * - remove rotation of v'(cartesian)
   * - copy back rotated v'(cartesian) into state->v
   */

  /* project velocities of index atoms into {es}. Either use 
   * wrapper project() or directly call project_to_eigvectors() 
   * since all evecs are moved to mon. 
   */
  project_to_eigvectors(loc->vdum,&edi->vecs.mon,edi,"v");  /* or: project(loc->vdum,edi,"v"); */

  /* construct v_es & v_nes */
  clear_rvecs(nr,v_es);         /* tidy up before calc.      */
  clear_rvecs(nr,v_nes);
  for (j=0; j<neig; j++)        /* do for all es evecs...    */
    for (i=0; i<nr; i++)        /* ...over all index atoms   */
      {
	v_es[i][XX] += edi->vecs.mon.vproj[j]*es_evec[j][i][XX];
	v_es[i][YY] += edi->vecs.mon.vproj[j]*es_evec[j][i][YY];
	v_es[i][ZZ] += edi->vecs.mon.vproj[j]*es_evec[j][i][ZZ];
      }
  for (i=0; i<nr; i++)          /* v_nes = v - v_es          */
    rvec_sub(loc->vdum[edi->sref.anrs[i]],v_es[i],v_nes[i]);
  
  /* calculate E_kin(t+dt/2) for proj. velocities => T_es(t+dt/2) & lambda_es(t+dt/2) for next step */ 
  ekin = 0.0;
  for (i=0; i<nr; i++)
    ekin += (md->massT[edi->sref.anrs[i]])*norm2(v_es[i]);
  ekin = 0.5*ekin;
  T = calc_temp(ekin,neig);
  if (T < 0)
    gmx_fatal(FARGS,"TEEREX  negative {es} temperature...I give up");

  /* T_es coupling (fac==-1) / v_es scaling after exchange (fac!=-1) */
  if (fac == -1)  /* T_es coupling */
    {
      dummy = 1 + dt/tau_es*(teerex->T_es[re->repl]/T - 1);     
      if (dummy < 0)
	gmx_fatal(FARGS,"TEEREX  lambda_es in Berendsen T_es-coupling negative!");
      teerex->lambda_es = sqrt(dummy); 

      /* Berendsen T_es-coupling:   v' = lambda_es_old * v     */
      for (i=0; i<nr; i++)
	svmul(teerex->lambda_es,v_es[i],v_es[i]);
    }
  else            /* v_es scaling after exchange */  
    {
      if (teerex->a == re->repl)
	fac = sqrt(teerex->T_es[teerex->a]/teerex->T_es[teerex->b]);
      else if (teerex->b == re->repl)
	fac = sqrt(teerex->T_es[teerex->b]/teerex->T_es[teerex->a]);
      for (i=0; i<nr; i++)
	svmul(fac,v_es[i],v_es[i]);
    }

  /* write out T_es(t) */
  if ((fac==-1) && do_per_step(step,ir->nstenergy))
    fprintf(log,"TEEREX:%i  step= %i     T_es= %f\n",re->repl,step,T);

  /* back-projection */
  for (i=0; i<nr; i++)
    rvec_add(v_es[i],v_nes[i],loc->vdum[edi->sref.anrs[i]]);

  /* remove velocity correction from correct_vel() */
  for (i=0; i<ned; i++)
    rvec_dec(loc->vdum[i],loc->vcorr[i]);

  /* rotate back scaled v */
  rmrotfit(ned,loc->vdum,rotmat);

  /* copy back loc->vdum into state->v */
  rveccopy(ned,loc->vdum,state->v);
  /*********************************************************/
  /* END        T_es coupling / velocity scaling       END */
  /*********************************************************/

  if (bFirst) 
    bFirst = FALSE;

  return;
}





